import sys

from demo.fixtures import (
    make_mappings,
    daphne_ds,
    desire_ds,
    populate_daphne,
    populate_desire,
    cleanup,
)
from sync.datasource import DataSource
from sync.syncer import Syncer

sys.path.append("src/")

import logging
from pprint import pformat


def main():
    logging.basicConfig(level=logging.INFO)
    logging.getLogger("sqlalchemy.engine").setLevel(level=logging.ERROR)

    cleanup()
    mappings = make_mappings()
    daphne_db = daphne_ds()
    desire_db = desire_ds()
    populate_daphne(daphne_db)
    populate_desire(desire_db)

    logging.info("Initial DB states")
    logging.info(show_db(desire_db))
    logging.info(show_db(daphne_db))

    syncer = Syncer(daphne_db, desire_db, mappings)
    operations = syncer.sync_all()
    logging.info(f"{len(operations)} operation(s) made")
    logging.info(f"Operations: {pformat(operations)}")

    logging.info("DB states after first and before second run")
    logging.info(show_db(desire_db))
    logging.info(show_db(daphne_db))

    operations = syncer.sync_all()
    logging.info(f"Sync with {len(operations)} operation(s) made")
    logging.info(f"Operations: {pformat(operations)}")

    logging.info("🙌 Final DB states")
    logging.info(show_db(desire_db))
    logging.info(show_db(daphne_db))

    operations = syncer.sync_all()
    logging.info(f"Sync with {len(operations)} operations made.")
    logging.info(f"Operations: {pformat(operations)}")
    if len(operations) == 0:
        logging.info("🎉🥳 Synchronization has converged.")
    else:
        logging.warning("Synchronization should have converged. Something is wrong.")


def show_db(ds: DataSource):
    return pformat(
        {
            entity_type: ds.all(entity_type, limit=10)
            for entity_type in ds.entity_types()
        }
    )


if __name__ == "__main__":
    main()



from abc import abstractmethod, ABC
from typing import List, Type


class DataSource(ABC):
    @abstractmethod
    def entity_types(self) -> List[type]:
        pass

    @abstractmethod
    def create(self, entity) -> bool:
        pass

    @abstractmethod
    def find(self, by_type, by_conditions) -> List:
        pass

    @abstractmethod
    def delete(self, entity) -> bool:
        pass

    @abstractmethod
    def all(self, of_type, limit=None) -> List:
        pass

    @abstractmethod
    def refresh(self, entity):
        pass

    @abstractmethod
    def preload_type(self, preload_type: Type, path: List[str]) -> bool:
        pass

    @abstractmethod
    def persist(self, entity):
        pass

    @abstractmethod
    def id(self, entity):
        pass

    @abstractmethod
    def get(self, type, id):
        pass



from demo.daphne import VlanServiceProvider
from demo.fixtures import daphne_sqa_ds, desire_sqa_ds, make_mappings
from sync import tasks

if __name__ == "__main__":
    ds = daphne_sqa_ds()
    vlsp = VlanServiceProvider(name="test")
    ds.create(vlsp)

    tasks.sync_one.delay(
        daphne_sqa_ds, desire_sqa_ds, make_mappings(), vlsp.__class__, vlsp.id
    )



import os

from sync.backends.inmemory import InMemoryDs
from sqlalchemy import Column, Integer, String, ForeignKey
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

from sync.backends.sqlalchemy import SQADataSource
from demo.daphne import DaphneBase, VlanServiceProvider, VlanServiceConnection
from demo.desire import (
    DesireBase,
    CloudServiceProvider,
    CloudRegion,
    CloudHandover,
    NIC,
)
from demo.util import Repr

from sync.mapping import Mapping, Mode, AttributeMap

DESIRE_SQLITE = "/tmp/desire.sqlite"
DAPHNE_SQLITE = "/tmp/daphne.sqlite"


def cleanup():
    if os.path.exists(DESIRE_SQLITE):
        os.remove(DESIRE_SQLITE)
    if os.path.exists(DAPHNE_SQLITE):
        os.remove(DAPHNE_SQLITE)


def daphne_ds():
    return InMemoryDs()


def daphne_sqa_ds():
    return SQADataSource(f"sqlite:///{DAPHNE_SQLITE}", DaphneBase)


def desire_ds():
    return InMemoryDs()


def desire_sqa_ds():
    return SQADataSource(f"sqlite:///{DESIRE_SQLITE}", DesireBase)


def populate_daphne(daphne):
    for o in {
        VlanServiceProvider(
            name="AWS",
            canUpgrade=False,
            children=[
                VlanServiceConnection(name="nic-1", region="eu-central-1", pop="INX6"),
                VlanServiceConnection(name="nic-2", region="eu-central-1", pop="EqFA5"),
                VlanServiceConnection(name="nic-3", region="eu-west-2", pop="EqFA5"),
            ],
        ),
        VlanServiceProvider(name="AZURE", canUpgrade=True, children=[]),
    }:
        daphne.persist(o)


def populate_desire(desire):
    for o in {
        CloudServiceProvider(
            name="AWS",
            children=[
                CloudRegion(
                    name="eu-central-1",
                    children=[
                        CloudHandover(name="INX6", children=[NIC(name="nic-1")]),
                        CloudHandover(name="EqFA5", children=[NIC(name="nic-2a")]),
                    ],
                ),
                CloudRegion(
                    name="eu-west-1",
                    children=[
                        CloudHandover(name="LON1", children=[NIC(name="nic-3a")])
                    ],
                ),
                CloudRegion(
                    name="eu-west-2",
                    children=[
                        CloudHandover(
                            name="EqFA5",
                            children=[NIC(name="nic-3", external_ref="nic3-eref")],
                        )
                    ],
                ),
            ],
        ),
        CloudServiceProvider(
            name="IBM",
            children=[
                CloudRegion(
                    name="EU-Frankfurt",
                    children=[
                        CloudHandover(name="fra03", children=[NIC(name="nic-4")])
                    ],
                )
            ],
        ),
    }:
        desire.persist(o)


def make_mappings():
    return [
        Mapping(
            entity_types=(VlanServiceConnection, CloudRegion),
            modes={Mode.LEFT_TO_RIGHT},
            keys={
                AttributeMap("region", "name"),
                AttributeMap(
                    "parent", "parent", VlanServiceProvider, CloudServiceProvider
                ),
            },
        ),
        Mapping(
            entity_types=(VlanServiceProvider, CloudServiceProvider),
            modes={Mode.LEFT_TO_RIGHT, Mode.RIGHT_TO_LEFT},
            keys={AttributeMap("name", "name")},
            attributes={AttributeMap("canUpgrade", "upgrade_allowed")},
        ),
        Mapping(
            entity_types=(VlanServiceConnection, CloudHandover),
            modes={Mode.LEFT_TO_RIGHT},
            keys={
                AttributeMap("pop", "name"),
                AttributeMap("__self__", "parent", VlanServiceConnection, CloudRegion),
            },
        ),
        Mapping(
            entity_types=(VlanServiceConnection, NIC),
            modes={Mode.RIGHT_TO_LEFT},
            keys={
                AttributeMap("name", "name"),
                AttributeMap("region", ("parent", "parent", "name")),
                AttributeMap("pop", ("parent", "name")),
                AttributeMap(
                    "parent",
                    ("parent", "parent", "parent"),
                    VlanServiceProvider,
                    CloudServiceProvider,
                ),
            },
            attributes={AttributeMap("externalRef", "external_ref")},
        ),
        Mapping(
            entity_types=(VlanServiceConnection, NIC),
            modes={Mode.LEFT_TO_RIGHT},
            keys={
                AttributeMap("name", "name"),
                AttributeMap(
                    "__self__", "parent", VlanServiceConnection, CloudHandover
                ),
            },
            attributes={AttributeMap("externalRef", "external_ref")},
        ),
    ]


UserCompanyBase = declarative_base()


class Company(UserCompanyBase, Repr):
    __tablename__ = "companies"
    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False)
    users = relationship("User", back_populates="company")


class User(UserCompanyBase, Repr):
    __tablename__ = "users"
    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False)
    company_id = Column(Integer, ForeignKey("companies.id"))
    company = relationship("Company", back_populates="users")



import os

from celery import Celery

from demo.fixtures import (
    make_mappings,
    populate_daphne,
    populate_desire,
    desire_sqa_ds,
    daphne_sqa_ds,
)

app = Celery(
    "synchronizer",
    broker="amqp://rabbitmq",
    backend="rpc://",
    include=["demo.celery.tasks"],
)

app.conf.update(result_expires=3600)
app.conf.task_serializer = "pickle"
app.conf.accept_content = ["pickle"]
app.conf.result_serializer = "pickle"
app.conf.result_accept_content = ["pickle"]

app.conf.beat_schedule = {
    "every-10-sec": {
        "schedule": 10,
        "task": "demo.celery.tasks.sync_all",
        "args": (daphne_sqa_ds, desire_sqa_ds, make_mappings()),
    }
}



from dataclasses import dataclass


@dataclass
class Addition:
    from_class: type
    to_class: type
    from_entity: object
    to_entity: object

    def __repr__(self):
        return f"Addition({self.to_entity})"


@dataclass
class Update:
    from_class: type
    to_class: type
    from_entity: object
    to_entity: object

    def __repr__(self):
        return f"Update({self.to_entity})"


@dataclass
class Linking:
    classes: frozenset
    entities: frozenset

    def __repr__(self):
        return f"Linking({self.entities})"


@dataclass
class SkipMissingLink:
    from_class: type
    to_class: type
    from_entity: object

    def __repr__(self):
        return (
            f"SkipMissingLink({self.from_class.__name__} -> {self.to_class.__name__})"
        )


@dataclass
class AttributeUpdated:
    entity: type
    attribute: str
    old_value: any
    new_value: any
    from_entity: object

    def __repr__(self):
        return f"AttributeUpdated({self.entity}.{self.attribute}: {self.old_value} -> {self.new_value} from {self.from_entity})"


@dataclass
class SkipUpdateOrphanedEntity:
    from_class: type
    to_class: type
    from_entity: object
    to_entity: object

    def __repr__(self):
        return f"SkipUpdateOrphanedEntity({self.from_class.__name__} -> {self.to_class.__name__}, {self.from_entity} -> {self.to_entity})"


@dataclass
class DeletedOrphanedEntity:
    from_class: type
    to_class: type
    from_entity: object
    to_entity: object

    def __repr__(self):
        return f"""
            DeletedOrphanedEntity(deleted {self.from_entity} of type {self.from_class}, 
            because {self.to_entity} of type {self.to_class} has disappeared and mapping
            allows deleting orphans)
        """


@dataclass
class CouldNotCreate:
    from_class: type
    to_class: type
    from_entity: object

    def __repr__(self):
        return f"CouldNotCreate({self.from_class.__name__} -> {self.to_class.__name__})"



import logging
import math
from dataclasses import dataclass
from typing import List, Set, Dict, Any, Type, Union, Tuple

from sync.backends.inmemory import InMemoryLinkProvider
from sync.datasource import DataSource
from sync.mapping import Mapping, Mode, AttributeMap, DirtyMode
from sync.mixins import Timestamped, LinkProvider
from sync.operations import (
    Addition,
    Linking,
    SkipMissingLink,
    AttributeUpdated,
    SkipUpdateOrphanedEntity,
    DeletedOrphanedEntity,
    CouldNotCreate,
)

MODES_TO_PARAMETERS = {
    Mode.LEFT_TO_RIGHT: ("from_attribute", "from_attribute_type_hint", 0),
    Mode.RIGHT_TO_LEFT: ("to_attribute", "to_attribute_type_hint", 1),
}


@dataclass
class Syncer:
    left_ds: DataSource
    right_ds: DataSource
    mappings: List[Mapping]
    link_provider: LinkProvider = InMemoryLinkProvider()

    def __post_init__(self):
        self._preload_data_sources()

    def sync_all(self):
        """
        Synchronizes :attribute:`left_ds` and :attribute:`right_ds` according to rules in :attribute:`mappings`

        :return: a list of operations that happened during synchronization for logging and debugging
        """
        operations = []
        for mapping in self.mappings:
            operations.extend(self.sync_mapping(mapping))
        return operations

    def sync_mapping(self, mapping: Mapping):
        operations = []

        if Mode.LEFT_TO_RIGHT in mapping.modes:
            logging.debug(
                f"Mapping from {mapping.entity_types[0].__name__} to {mapping.entity_types[1].__name__}"
            )
            operations.extend(self._sync_from_to(self.left_ds, self.right_ds, mapping))
        if Mode.RIGHT_TO_LEFT in mapping.modes:
            logging.debug(
                f"Mapping from {mapping.entity_types[1].__name__} to {mapping.entity_types[0].__name__}"
            )
            operations.extend(
                self._sync_from_to(self.right_ds, self.left_ds, mapping.swapped())
            )
        return operations

    def _preload_data_sources(self):
        def preload_data_source_type(attribute, data_source, hint, index, key, mapping):
            path = getattr(key, attribute)
            if isinstance(path, tuple) or (
                getattr(key, hint) is not None and path != "__self__"
            ):
                entity_type = mapping.entity_types[index]
                if not isinstance(path, tuple):
                    path = (path,)
                logging.debug(
                    f"Preloading from {entity_type} path {path} from {data_source}"
                )
                data_source.preload_type(entity_type, path)

        for mode in Mode:
            for mapping in self.mappings:
                for key in mapping.keys:
                    attribute, hint, index = MODES_TO_PARAMETERS[mode]
                    data_source = [self.left_ds, self.right_ds][index]
                    preload_data_source_type(
                        attribute, data_source, hint, index, key, mapping
                    )

    def _upsert_unlinked_entity(
        self,
        from_ds: DataSource,
        from_entity,
        mapping: Mapping,
        data_source: DataSource,
        entity_type: Type,
    ):
        operations = []
        try:
            conditions = self._create_key_value_pairs(
                from_entity, mapping.keys, from_ds=from_ds, to_ds=data_source,
            )
            found_entities = data_source.find(entity_type, conditions)
        except NotLinked as e:
            operations.append(
                SkipMissingLink(from_entity.__class__, entity_type, from_entity)
            )
            logging.warning(e)
            return operations

        if len(found_entities) == 0:
            from_entity_timestamp = self._timestamp(from_entity)

            try:
                non_key_attributes = self._create_key_value_pairs(
                    from_entity, mapping.attributes, from_ds=from_ds, to_ds=data_source,
                )
                to_entity = Syncer._create_entity(
                    conditions, non_key_attributes, entity_type, from_entity_timestamp
                )
            except NotLinked as e:
                operations.append(
                    SkipMissingLink(from_entity.__class__, entity_type, from_entity)
                )
                logging.warning(e)
                return operations

            if not mapping.create or not data_source.create(to_entity):
                logging.warning(f"Could not add {to_entity.__repr__()}")
                operations.append(
                    CouldNotCreate(from_entity.__class__, entity_type, from_entity)
                )
                return operations

            if mapping.dirty_mode == DirtyMode.UPDATE:
                setattr(from_entity, mapping.dirty_attribute_name, False)
                logging.info(
                    f"Removed {mapping.dirty_attribute_name} flag from {from_entity}"
                )
                from_ds.persist(from_entity)

            operations.append(
                Addition(from_entity.__class__, entity_type, from_entity, to_entity)
            )
            logging.info(f"✨ Adding {to_entity.__repr__()}")

            if self.link_provider.link(to_entity, from_entity, data_source, from_ds):
                classes = frozenset(mapping.entity_types)
                entities = frozenset({from_entity, to_entity})
                operations.append(Linking(classes, entities))
        elif len(found_entities) == 1:
            to_entity = next(iter(found_entities))

            if self.link_provider.link(to_entity, from_entity, data_source, from_ds):
                classes = frozenset(mapping.entity_types)
                entities = frozenset({from_entity, to_entity})
                operations.append(Linking(classes, entities))
                operations.extend(
                    self._update_linked_entity(
                        from_ds,
                        from_entity,
                        {to_entity},
                        mapping,
                        to_entity.__class__,
                        data_source,
                    )
                )
        else:
            logging.warning(f"Found more than one result for {conditions}, skipping")

        return operations

    @staticmethod
    def _timestamp(entity) -> int:
        return entity.updated_at() if isinstance(entity, Timestamped) else 0

    @staticmethod
    def _collect_objects_in_paths(entity, paths) -> int:
        entities = set()
        for path in paths:
            entities = entities.union(Syncer._collect_objects_in_path(entity, path))
        logging.debug(f"Found objects in paths of {entity} paths {paths}: {entities}")
        return entities

    @staticmethod
    def _resolve_entity(entity, key):
        if key != "__self__":
            return getattr(entity, key)
        else:
            return entity

    @staticmethod
    def _collect_objects_in_path(entity, path):
        entities = {entity}

        if isinstance(path, tuple) or isinstance(path, list):
            if len(path) > 1:
                key, *rest = path
                next_entity = Syncer._resolve_entity(entity, key)
                if not next_entity:
                    return entities
                entities = entities.union(
                    Syncer._collect_objects_in_path(next_entity, rest)
                )
            elif len(path) == 1:
                key = path[0]
                entities.add(Syncer._resolve_entity(entity, key))
        elif isinstance(path, str):
            entities.add(Syncer._resolve_entity(entity, path))

        return entities

    @staticmethod
    def _biggest_timestamp(entities: List) -> int:
        max_value = 0
        for entity in entities:
            max_value = max(max_value, Syncer._timestamp(entity))
        return max_value

    @staticmethod
    def _set_updated_at(entity: Timestamped, timestamp: int):
        if isinstance(entity, Timestamped):
            logging.debug(f"Setting timestamp of {entity} to {timestamp}")
            entity.update_timestamp(timestamp)

    @staticmethod
    def _is_dirty(mapping, entity):
        try:
            return getattr(entity, mapping.dirty_attribute_name)
        except AttributeError as e:
            logging.warning(
                f"asked to read {mapping.dirty_attribute_name} from {entity} but no such attribute exists"
            )
            raise e

    def _update_linked_entity(
        self,
        from_ds: DataSource,
        from_entity,
        others: List,
        mapping: Mapping,
        to_class: Type,
        to_ds: DataSource,
    ):
        operations = []

        try:
            attributes = {
                **self._create_key_value_pairs(
                    from_entity, mapping.keys, from_ds=from_ds, to_ds=to_ds,
                ),
                **self._create_key_value_pairs(
                    from_entity, mapping.attributes, from_ds=from_ds, to_ds=to_ds,
                ),
            }
        except NotLinked as e:
            logging.warning(e)
            operations.append(
                SkipMissingLink(from_entity.__class__, to_class, from_entity)
            )
            return operations

        # TODO more tests, maybe validate modes
        for to_entity in others:
            refreshed_to_entity = to_ds.refresh(to_entity)

            if not refreshed_to_entity:
                logging.warning(
                    f"Orphaned entities in link from {from_entity}: {refreshed_to_entity}"
                )
                operations.append(
                    SkipUpdateOrphanedEntity(
                        from_entity.__class__, to_class, from_entity, to_entity
                    )
                )

                if mapping.delete_orphaned_entities:
                    if from_ds.delete(from_entity):
                        self.link_provider.unlink(
                            from_entity, entity_ds=from_ds, other_ds=to_ds
                        )
                        operations.append(
                            DeletedOrphanedEntity(
                                from_entity.__class__, to_class, from_entity, to_entity
                            )
                        )
                continue
            else:
                to_entity = refreshed_to_entity

            if mapping.dirty_mode == DirtyMode.DO_NOT_OVERWRITE and self._is_dirty(
                mapping, to_entity
            ):
                continue

            # guard that to_entity has last been changed before or at the same time as from_entity
            from_entity_objects = Syncer._collect_objects_in_paths(
                from_entity, mapping.from_paths()
            )
            timestamp_from_entity = Syncer._biggest_timestamp(from_entity_objects)
            timestamp_to_entity = Syncer._timestamp(to_entity)
            if timestamp_to_entity > timestamp_from_entity:
                logging.warning(
                    f"Not updating {to_entity}: is newer than {from_entity}"
                )
                continue

            entity_updated = False
            for k, v in attributes.items():
                to_value = getattr(to_entity, k)
                if to_value != v:
                    logging.info(f"👌 Updating {to_entity}.{k} {to_value} -> {v}")
                    setattr(to_entity, k, v)
                    entity_updated = True

                    operations.append(
                        AttributeUpdated(
                            entity=to_entity,
                            attribute=k,
                            old_value=to_value,
                            new_value=v,
                            from_entity=from_entity,
                        )
                    )

            if timestamp_to_entity != timestamp_from_entity:
                self._set_updated_at(to_entity, timestamp_from_entity)
                entity_updated = True
            if entity_updated:
                to_ds.persist(to_entity)
                if mapping.dirty_mode == DirtyMode.UPDATE:
                    setattr(from_entity, mapping.dirty_attribute_name, False)
                    logging.info(
                        f"Removed {mapping.dirty_attribute_name} flag from {from_entity}"
                    )
                    from_ds.persist(from_entity)

        return operations

    def _sync_from_to(
        self, from_ds: DataSource, to_ds: DataSource, mapping: Mapping,
    ):
        operations = []
        from_class, to_class = mapping.entity_types
        from_entities: List = from_ds.all(from_class)

        for from_entity in from_entities:
            operations.extend(
                self._sync_entity(from_ds, from_entity, to_ds, to_class, mapping)
            )

        return operations

    def sync_entity(self, type, id):
        operations = []
        if type in self.left_ds.entity_types():
            from_ds = self.left_ds
            to_ds = self.right_ds
        elif type in self.right_ds.entity_types():
            from_ds = self.right_ds
            to_ds = self.left_ds
        else:
            raise AttributeError(f"Unknown entity type: {type.__name__}")

        from_entity = from_ds.get(type, id)

        for mapping in self.mappings:
            entity_types = mapping.entity_types
            if from_entity.__class__ in entity_types:
                to_class_index = 1 - entity_types.index(from_entity.__class__)
                operations.extend(
                    self._sync_entity(
                        from_ds,
                        from_entity,
                        to_ds,
                        entity_types[to_class_index],
                        mapping,
                    )
                )

        return operations

    def _sync_entity(self, from_ds, from_entity, to_ds, to_class, mapping):
        operations = []

        if mapping.dirty_mode == DirtyMode.UPDATE and not self._is_dirty(
            mapping, from_entity
        ):
            return operations

        others = self.link_provider.others(from_entity, from_ds, to_class, to_ds)
        if not others or not len(others):
            operations.extend(
                self._upsert_unlinked_entity(
                    from_ds, from_entity, mapping, to_ds, to_class
                )
            )
        else:
            operations.extend(
                self._update_linked_entity(
                    from_ds, from_entity, others, mapping, to_class, to_ds,
                )
            )
        return operations

    @staticmethod
    def _update_entity(entity, **kwargs):
        for (key, value) in kwargs.items():
            setattr(entity, key, value)

    @staticmethod
    def _create_entity(
        conditions: Dict[str, Any],
        non_key_attributes: Dict[str, Any],
        to_class: Type,
        timestamp: int = None,
    ):
        to_entity = to_class()
        Syncer._update_entity(to_entity, **conditions)
        Syncer._update_entity(to_entity, **non_key_attributes)
        if timestamp:
            to_entity.update_timestamp(timestamp)
        return to_entity

    def _create_key_value_pairs(
        self,
        from_entity,
        attribute_maps: Set[AttributeMap],
        from_ds: DataSource,
        to_ds: DataSource,
    ):
        attributes = {}

        for attribute_map in attribute_maps:
            value, linked_value = self._resolve_values(
                from_entity,
                attribute_map.from_attribute,
                attribute_map.from_attribute_type_hint,
                attribute_map.to_attribute_type_hint,
                entity_ds=from_ds,
                other_ds=to_ds,
            )
            if not isinstance(attribute_map.to_attribute, str):
                raise MayNotUpdateJoinedAttributes(
                    f"May not update attributes {attribute_map.to_attribute} of {attribute_map.to_attribute_type_hint}"
                )
            attributes[attribute_map.to_attribute] = linked_value or value

        logging.debug(f"attributes: {attributes}")
        return attributes

    def _resolve_values(
        self,
        entity,
        key: Union[str, Tuple],
        from_remote_class: Type = None,
        to_remote_class: Type = None,
        entity_ds: DataSource = None,
        other_ds: DataSource = None,
    ):
        linked_value = None

        if key == "__self__":
            value = entity
        else:
            value = Syncer._resolve_key_value(entity, key)

        #        if value is None:
        #            raise UnableToResolve(f"Cannot resolve {entity}.{key}")

        # If type hint is provided, try to resolve value type via typed link
        if from_remote_class and value:
            if not self.link_provider:
                raise NotLinked(f"No link provider provided")
            linked_values = self.link_provider.others(
                value, entity_ds, to_remote_class, other_ds
            )
            if not linked_values:  # None or empty set
                raise NotLinked(
                    f"Cannot find linked entity of {entity}.{key}. {value} not in {to_remote_class.__name__}"
                )
            if len(linked_values) > 1:
                raise MultiLinkedEntity(
                    f"Cannot resolve a single value from multi-linked entity"
                )
            linked_value = linked_values.pop()

        logging.debug(f"resolve_value({entity},{key})={value},{linked_value}")

        return value, linked_value

    @staticmethod
    def _resolve_key_value(entity, key: str):
        if isinstance(key, tuple) or isinstance(key, list):
            if len(key) > 1:
                next_key, *rest = key
                next_entity = getattr(entity, next_key)
                if not next_entity:
                    return None
                return Syncer._resolve_key_value(next_entity, rest)
            elif len(key) == 1:
                key = key[0]
            else:
                raise AttributeError("key is empty")
        return getattr(entity, key)


class NotLinked(Exception):
    pass


class UnableToResolve(Exception):
    pass


class MultiLinkedEntity(Exception):
    pass


class MayNotUpdateJoinedAttributes(Exception):
    pass



import logging
from typing import List, Type

from sqlalchemy import create_engine
from sqlalchemy import exc
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, Session, joinedload

from sync.datasource import DataSource

Session = sessionmaker(expire_on_commit=False)


class JoinedLoads:
    def __init__(self):
        self.joinedloads = dict()

    def add(self, type, jl):
        if not type in self.joinedloads:
            self.joinedloads[type] = set()
        self.joinedloads[type].add(jl)

    def get(self, type):
        return self.joinedloads[type] if type in self.joinedloads else set()


class SQADataSource(DataSource):
    def __init__(
        self,
        url="sqlite+pysqlite:///testdb.sqlite",
        base: Type = declarative_base(),
        create_metadata=True,
    ):
        self.url = url
        self.base = base
        self.engine = create_engine(url, echo=False)
        Session.configure(bind=self.engine)
        if create_metadata:
            base.metadata.create_all(self.engine)
        self.session: Session = Session()
        self.joinedloads = JoinedLoads()

    def create(self, entity: object) -> bool:
        return self.persist(entity)

    def find(self, by_type: Type, by_conditions) -> List[object]:
        query = self.session.query(by_type)

        for k, v in by_conditions.items():
            query = query.filter(getattr(by_type, k) == v)

        for joinedload in self.joinedloads.get(by_type):
            query = query.options(joinedload)

        return query.all()

    def delete(self, entity) -> bool:
        self.session.delete(entity)
        return self.session.commit()

    def all(self, of_type, limit=None) -> set:
        return self.session.query(of_type).limit(limit).all()

    def refresh(self, entity) -> object:
        if entity:
            logging.debug(f"Refreshing {entity} on session: {self.session}")
            self.session.refresh(entity)
        return entity

    def preload_type(self, preload_type, path) -> bool:
        jl = None
        current_type = preload_type
        for prop in path:
            if SQADataSource._is_relationship(current_type, prop):
                rel = getattr(current_type, prop)
                if not jl:
                    jl = joinedload(rel)
                else:
                    jl = jl.joinedload(rel)
                current_type = SQADataSource._resolve_remote_class(current_type, prop)
            else:
                break
        self.joinedloads.add(preload_type, jl)

    def entity_types(self) -> List[type]:
        return self.base.__subclasses__()

    def persist(self, entity) -> object:
        logging.debug(f"Persisting {entity}")
        self.session.add(entity)
        try:
            self.session.commit()
            return True
        except exc.SQLAlchemyError as err:
            logging.exception(msg="Cannot persist", exc_info=err)
            self.session.rollback()
            return None

    @staticmethod
    def _resolve_remote_class(type_, attribute):
        return type_.__mapper__.relationships.get(attribute).entity.class_

    @staticmethod
    def _is_relationship(type_, attribute):
        return attribute in type_.__mapper__.relationships

    def id(self, entity):
        return entity.id

    def get(self, type, id):
        return self.session.query(type).get(id)

    def __repr__(self):
        return f"SQADataSource(base={self.base}, url={self.url})"
